import pygame
from pygame import Surface
import math
from pygame.locals import *
from symbol import Symbol, ButtonSymbol
from pytbot.dumbo import TBot
import sys
from buttons import Button as Btn
from colors import random_color
import random
from deuces import Card, Evaluator
import re
import time
from pygame.mixer import Sound

RESOLUTION_X = 1200
RESOLUTION_Y = 800
WHITE = [255,255,255,255]
symbol_cache = {}
CHANNEL = "#nowhere" #shhhh don't pay attention
CREG = "([2-9TJQKA][hdsc])"

START_CASH = 1000 + random.randint(-500, 500)

pygame.init()
pygame.mixer.init()
SCREEN = pygame.display.set_mode((RESOLUTION_X, RESOLUTION_Y), pygame.SRCALPHA)
MISC_FONT = pygame.font.Font('FreeSerif.ttf', 13)
GAMEOVER_SYMBOL = Symbol([200,25,25], 5)
GAMEOVER_SURF = Surface((RESOLUTION_X, RESOLUTION_Y)).convert_alpha()
GAMEOVER_SURF.fill([1,1,1,150])
GAMEOVER_SYMBOL.draw(GAMEOVER_SURF, (100,100))
YOUWIN_SURF = Surface((RESOLUTION_X, RESOLUTION_Y)).convert_alpha()
YOUWIN_SYMBOL = Symbol([25,200,25], 5)
POT_SYMBOL = Symbol([25,25,85], 3)
FACE_SOURCE = pygame.image.load('reticulan_portraits_recolours.png').convert_alpha()
FACE_SOURCE = pygame.transform.scale2x(FACE_SOURCE)
def color_clamp(threshold):
    return [random.randint(0,threshold) for _ in xrange(1,4)]

def get_card_symbol(value):
    if value not in symbol_cache:

        new_s = Symbol([0,0,0])
        while new_s.letters_used[0] in [v.letters_used[0] for v in symbol_cache.values()]:
            new_s = Symbol([0,0,0])
        symbol_cache[value] = new_s   
    return symbol_cache[value]

suitcolor_cache = {}

def get_color(suit):
    if suit not in suitcolor_cache:
        suitcolor_cache[suit] = color_clamp(190)
    return suitcolor_cache[suit]

card_face_cache = {}

def get_card_face(val, suit, width, height):
    if val+suit not in card_face_cache:
        surf = Surface((width,height)).convert_alpha()
        surf.fill([255,255,255,0])
        if val in ['K', 'Q', 'J']:
            a = (128 * random.randint(0,4), 128 * random.randint(0,4), 128,128)
            surf.blit(FACE_SOURCE, (0,0), area=a)
            tmp = Surface((width, 128)).convert_alpha()
            tmp.fill([255,255,255,0])
            tmp.blit(FACE_SOURCE, (0,0), area=a)
            tmp = pygame.transform.rotate(tmp, 180)
            surf.blit(tmp, (0,height-128, width, height/2))
            #return surf
        
        elif val == 'A':
            sym = get_card_symbol(suit)
            sym.draw(surf, (width/2, height/2))
            #
        else:
            times = 0
            if val == 'T':
                times = 10
            else:
                times = int(val)
            sym = get_card_symbol(suit)
            for i in xrange(0, times):
                sym.draw(surf, (random.randint(0, width-sym.surf.get_height()), random.randint(0,height-sym.surf.get_width())))

        card_face_cache[str(val)+suit] = surf

    return card_face_cache[str(val)+suit]


music = ["sounds/m"+str(x)+".ogg" for x in xrange(1,11)]

random.shuffle(music)
pygame.mixer.music.load(music[0])
pygame.mixer.music.set_volume(.7)
pygame.mixer.music.play()


class Point(object):
    def __init__(self, x, y):
        self.x = x
        self.y = y

class Main(object):

    def __init__(self, screen):
        self.screen = screen
        self.main_loop = True
        self.setup_poker()
        self.setup_buttons()
        self.player_action = False
        self.bg = color_clamp(80)
        self.displaycards = []
        self.board = []
        self.player_to_act = False
        self.to_call = 0
        self.hand_ending = False
        self.flopped = False
        self.turned = False
        self.rivered = False
        self.hand_will_end = False
        self.gameover = False
        self.youwin = False
        self.wait_time = 0
        self.pot = 0
        self.sounds = {}
        self.playsound = False
        self.playsound2 = False
        self.last_progress = 0


    def update(self, dt, time):
        pygame.event.pump()
        pressed = pygame.key.get_pressed()
        if pressed[K_ESCAPE]:
            self.main_loop = False

        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()
            elif event.type == MOUSEBUTTONDOWN:
                pos = pygame.mouse.get_pos()
                if self.check_btn.pressed(pos):
                    #print  "check"
                    self.player_action = "p check"

                if self.call_btn.pressed(pos):
                    #print  "call"
                    self.player_action = "p call"

                if self.raise_btn.pressed(pos):
                    #print  "raise"
                    self.player_action = "p raise 75"

                if self.allin_btn.pressed(pos):
                    #print  "allin"
                    self.player_action = "p jam"

                if self.fold_btn.pressed(pos):
                    #print  "fold"
                    self.player.folded = True
                    self.player.render()
                    self.player_action = "p fold"

        for b in self.bots:
            b.draw(self.screen)

        self.player.draw(self.screen)

        for i,card in enumerate(self.displaycards):
            card.draw(self.screen, (325 + 120 * i, 160 + 10 * i))

        x,y = 260,360
        self.screen.fill([235,235,255],(x, y, 120, 70))
        self.screen.blit(POT_SYMBOL.get_surf(), (x+10, y+10))
        self.screen.blit(MISC_FONT.render(str(self.pot), True, [60,60,60]), (x+10, y+45))

        if self.player_to_act == 'player':
            self.check_btn.create_button()
            self.raise_btn.create_button()
            self.allin_btn.create_button()
            self.call_btn.create_button()
            self.fold_btn.create_button()

        if self.hand_will_end == True :
            for b in self.bots:
                for card in b.displaycards:
                    card.facedown = False
                    card.render()
                b.render()
                b.folded = False
            if not b.folded:
                b.draw(self.screen)

            self.hand_will_end = False

            self.board = []
            self.displaycards = []
            for b in self.bots:
                for card in b.displaycards:
                    card.facedown = True
                    card.render()
                b.render()
            self.player.render()

            self.end_hand()
            

        if not self.waiting:
            self.progress(dt, time)
        
        if self.playsound2:
            self.play_sound
            self.playsound2 = False
            self.playsound = True

        if self.playsound:
            self.play_sound()
            self.playsound = False

        ##print  symbol_cache

    def progress(self, dt, time):
        if self.gameover:
            sys.exit()

        if self.youwin:
            sys.exit()
            

        if self.player_to_act == 'player':
            if self.player_action != False:
                self.player.msg(self.player_action)
                self.player_action = False

        elif self.player_to_act and self.player_to_act[:3] == 'bot':
            bot = self.bots[int(self.player_to_act[3])]
            bot.act(self.to_call)

        self.last_progress = time

        

    def setup_poker(self):
        self.p = TBot(self)
        self.p.joined(CHANNEL) # don't look at this!!!
        self.bots = []

    def play_sound(self):
        index_1 = random.randint(1,2)
        index_2 = random.randint(1,5)
        name = str(index_1) + str(index_2) + '.ogg'
        fname = 'sounds/'+name

        s = self.get_sound(fname)

        s.play()


    def get_sound(self, fname):
        if fname not in self.sounds:
            self.sounds[fname] = Sound(fname)
        return self.sounds[fname]


    def setup_buttons(self):
        r = Symbol([60,60,60], length=5)
        self.check_btn = Btn(self.screen, (107,142,35), 265, 640, 130, 45, 1, r)
        r = Symbol([60,60,60], length=4)
        self.call_btn = Btn(self.screen, (107,35,147), 265, 565, 130, 45, 1, r)
        r = Symbol([60,60,60], length=1)
        self.raise_btn = Btn(self.screen, (147,35,107), 265, 490, 50, 45, 1, r)
        r = Symbol([60,60,60], length=2)
        self.allin_btn = Btn(self.screen, (35,107,142), 345, 490, 50, 45, 1, r)
        r = Symbol([60,60,60], length=4)
        self.fold_btn = Btn(self.screen, (35,142,107), 265, 715, 130, 45, 1, r)

    def run(self):
        self.screen.convert_alpha()

        FPS = 60
        time = 0
        clock = pygame.time.Clock()

        curr_music = 0

        self.start()

        while self.main_loop:

            milliseconds = clock.tick(FPS)
            dt = milliseconds / 100.0
            time += dt
            if self.wait_time < 0:
                self.wait_time = 0
            if self.wait_time > 0:
                self.wait_time -= dt/3
            if self.wait_time == 0:
                self.waiting = False
            else:
                self.waiting = True

            ##print  time - self.last_progress
            if time - self.last_progress > 200:
                self.player.msg('??')
                self.player.msg('p fold')

            if random.random() > .95:
                self.player.msg('keepalive')


            self.screen.fill(self.bg)
            self.update(dt, time)
            if not pygame.mixer.music.get_busy() and not self.gameover:
                curr_music = curr_music + 1
                if curr_music > len(music):
                    curr_music = 0
                pygame.mixer.music.load(music[curr_music])
                pygame.mixer.music.play()

            if self.gameover:
                self.screen.blit(GAMEOVER_SURF, (0,0))
                pygame.mixer.music.fadeout(3)
                self.player.draw(self.screen)

            if self.youwin:
                pygame.mixer.music.set_volume(1.0)
                YOUWIN_SYMBOL.draw(YOUWIN_SURF, (random.randint(0, RESOLUTION_X), random.randint(0, RESOLUTION_Y)))
                self.screen.blit(YOUWIN_SURF, (0,0))


            pygame.display.set_caption("?")
            pygame.display.flip()

        pygame.quit()
        sys.exit()


    def start(self):
        for i in xrange(0,3):
            self.bots.append(Bot(self, i))
            self.bots[i].join()


        self.player = Player(self)
        self.player.join()

        self.player.start()

    def end_hand(self):
        for b in self.bots:
            b.end_hand()
            #print  b.name, 'has', b.cash
        #self.board =[]
        #self.displaycards = []
        self.flopped = False
        self.turned = False
        self.rivered = False
        self.hand_ending = False
        #self.player_to_act = False
        #self.to_call = 0



    def get_bot(self, pitard):
        return self.bots[int(pitard[3])]

    def is_bot(self, pitard):
        return pitard[:3] == "bot"

    def hand_over(self, match):
        #print  'hand over'
        p = match.group(1)
        #print  p, 'wins'
        if p == 'player':
            self.player.cash += int(match.group(2))
        for b in self.bots:
            if b.name == p:
                b.cash += int(match.group(2))
            if not b.folded:
                for card in b.displaycards:
                    card.facedown = False
                    card.render()
                b.render()
                b.draw(self.screen)
            b.folded = False
        #self.update(0)
        self.wait_time = 15
        self.hand_will_end = True

    def handle_public(self, msg):
        #print  'GAME handling', msg

        actmatch = re.match('^([playerbot012]+) is next to act. \((\d+) to call\)$', msg)
        if actmatch:
            #print  'act' + actmatch.group(1)
            self.player_to_act = actmatch.group(1)
            self.to_call = int(actmatch.group(2))
            self.wait_time = 10

            for b in self.bots:

                b.render()
            self.player.render()

        blindmatch = re.match('^([playerbot012]+) blinds \$(\d+)\.', msg)
        if blindmatch:
            #print  'bind'
            p = blindmatch.group(1)
            amt = int(blindmatch.group(2))
            if p == 'player':
                self.player.cash -= amt
            for b in self.bots:
                if p == b.name:
                    b.cash -= amt
            self.wait_time = 10

        foldmatch = re.match('^([playerbot012]+) folds\..+', msg)
        if foldmatch:
            p = foldmatch.group(1)
            #print  p, "folded"
            if p == 'player':
                self.player.folded = True
            for b in self.bots:
                if p == b.name:
                    b.folded = True
                    self.playsound = True
            self.wait_time = 10

        raisematch = re.match('^([playerbot012]+)\s+raises\s+\$(\d+)[^\$]+\$(\d+)', msg)
        if raisematch:
            p = raisematch.group(1)
            amt = int(raisematch.group(2))
            self.pot = int(raisematch.group(3))
            #print  'raise'
            if p == 'player':
                self.player.cash -= amt
            for b in self.bots:
                if p == b.name:
                    b.cash -= amt
                    self.playsound = True
            self.wait_time = 10

        betsmatch = re.match('^([playerbot012]+)\s+bets\s+\$(\d+).*Pot is now \$(\d+)', msg)
        if betsmatch:
            p = betsmatch.group(1)
            amt = int(betsmatch.group(2))
            self.pot = int(betsmatch.group(3))
            #print  'bet'
            if p == 'player':
                self.player.cash -= amt
            for b in self.bots:
                if p == b.name:
                    b.cash -= amt
                    self.playsound = True
            self.wait_time = 10

        handovermatch = re.match('^[A-Za-z]+: ([a-z0-9]+) wins \$(\d+) wi.+$', msg)
        if handovermatch:
            self.hand_over(handovermatch)

        handovermatch2 = re.match('^([playerbot012]+)\s+wins\s+\$(\d+)', msg)
        if handovermatch2:
            self.hand_over(handovermatch2)

        callmatch = re.match('^([playerbot012]+)\s+calls\s+\$(\d+)[^\$]+\$(\d+)', msg)
        if callmatch:
            p = callmatch.group(1)
            amt = int(callmatch.group(2))
            self.pot = int(callmatch.group(3))
            #print  p, 'calls', amt

            if p == 'player':
                self.player.cash -= amt
            for b in self.bots:
                if p == b.name:
                    b.cash -= amt
            self.wait_time = 10

        showhandsmatch = re.match("^([playerbot012]+)\s+:\s%s\s%s$" %(CREG, CREG), msg)
        if showhandsmatch:
            p = showhandsmatch.group(1)

            if self.is_bot(p):
                for card in self.bots[int(p[3])].displaycards:
                    card.facedown = False
                    card.render()
                    self.wait_time = 15

        flopmatch = re.match("^Board:\s+%s\s%s\s%s\s+$" % (CREG, CREG, CREG), msg)
        if flopmatch:
            #print  'flop'
            self.board = [flopmatch.group(1), flopmatch.group(2), flopmatch.group(3)]
            for card in self.board:
                self.displaycards.append(DisplayCard(card[1], card[0], facedown=False))
            self.flopped = True
            self.wait_time = 10

        turnmatch = re.match("^Board:\s+%s\s%s\s%s\s%s\s+$" % (CREG, CREG, CREG, CREG), msg)
        if turnmatch:
            #print  'turn'
            self.board.append(turnmatch.group(4))
            self.displaycards.append(DisplayCard(self.board[-1][1], self.board[-1][0], facedown=False))
            self.displaycards[-1].render()
            self.turned = True
            self.wait_time = 10

        rivermatch = re.match("^Board:\s+%s\s%s\s%s\s%s\s%s\s+?$"% (CREG, CREG, CREG, CREG, CREG), msg)
        if rivermatch:
            #print  'river'
            self.board.append(rivermatch.group(5))
            self.displaycards.append(DisplayCard(self.board[-1][1], self.board[-1][0], facedown=False))
            self.rivered = True
            self.wait_time = 10

        bustedmatch = re.match("^([playerbot012]+)\s+is\s+busted\!", msg)
        if bustedmatch:
            #print  "busted"
            p = bustedmatch.group(1)
            if p == 'player':
                self.player.busted = True
                self.player.cash = 0
                self.gameover = True
                self.wait_time = 30
                playsound2 = True
            else:
                self.get_bot(p).busted = True
                self.wate_time = 15
                self.playsound = True

        winmatch = re.match("^The tourney is over\.\s+([playerbot012]+)", msg)
        if winmatch:
            p = winmatch.group(1)
            if p == 'player':
                self.youwin = True
                self.wait_time = 30
            else:
                self.playsound = True



class DisplayCard(object):
    h = 281
    w = 200
    def __init__(self, suit, value, facedown=True):
        self.suit = suit
        self.value = value
        self.surf = Surface((DisplayCard.w,DisplayCard.h)).convert_alpha()

        self.facedown = facedown
        self.render()

    def render(self):
        if self.facedown:
            self.surf = pygame.image.load('back.png')
        else:
            self.surf.fill([0,0,0])
            self.surf.fill([255,255,255], rect=(1,1,DisplayCard.w-1, DisplayCard.h-1))
            s = get_card_symbol(self.suit)
            s.color = get_color(self.suit)
            v = get_card_symbol(self.value)
            v.color = s.color
            h = s.surf.get_height() + 50 +v.surf.get_height()
            w = max(s.surf.get_width(), v.surf.get_width())
            tmp = Surface((w,h)).convert_alpha()
            tmp.fill(WHITE)
            v.draw(tmp, ((w-v.surf.get_width())/2,0))
            y = 50
            if (s.letters[0][0] == 0):
                y = 40
            s.draw(tmp, ((w-s.surf.get_width())/2,y))
            self.surf.blit(tmp, (10,10))
            reverse_x = DisplayCard.w - tmp.get_width()
            reverse_y = DisplayCard.h - tmp.get_height()  
            tmp = pygame.transform.rotate(tmp, 180)          
            self.surf.blit(tmp, (reverse_x-10, reverse_y-10))

            self.surf.fill([0,0,0,255], (50, 20, DisplayCard.w - 100, DisplayCard.h - 40))
            self.surf.fill([255,255,255,255], (51, 21, DisplayCard.w - 102, DisplayCard.h - 42))
            face = get_card_face(self.value, self.suit, DisplayCard.w - 102, DisplayCard.h - 42) 
            self.surf.blit(face, (51,21))


    def draw(self, layer, pos):
        #self.render()
        layer.blit(self.surf, pos)



class Bot(object):
    def __init__(self, game, index, name=False):
        if name == False:
            self.name = "bot"+str(index)
        else:
            self.name = name
        self.id = index
        self.dumbo = game.p
        self.game = game
        self.cash = START_CASH
        self.curr_bet = 0
        self.pot = 0
        self.hand = []
        self.ranker = Evaluator()
        self.curr_rank = 8000
        self.displaycards = []
        self.folded = False
        self.surf = Surface((DisplayCard.w + 55, DisplayCard.h+35)).convert_alpha()
        self.nameplate = Symbol([60,60,60], random.randint(1,3)).get_surf()
        self.busted = False

    def end_hand(self):
        self.pot = 0
        self.curr_bet = 0
        self.folded = False
        #self.hand = []
        #self.displaycards = []
        self.render()

    def msg(self, msg):
        #print  self.name, msg
        self.dumbo.privmsg(self.name, CHANNEL, msg)

    def join(self):
        self.dumbo.fake_players[self.name] = self
        self.msg('p join')

    def handle(self, msg):
        #print  self.name, "handling", msg

        cardsmatch = re.match('^Your hole cards are: %s %s$' % (CREG, CREG), msg)
        if cardsmatch:
            self.hand = [cardsmatch.group(1), cardsmatch.group(2)]
            #print  self.name, 'got cards', self.hand
            self.displaycards = []
            for card in self.hand:
                self.displaycards.append(DisplayCard(card[1], card[0]))
            self.folded = False

        callorfoldmatch = re.match('Please call or fold', msg)
        if callorfoldmatch:
            if random.random() > .6:
                self.msg('p call')
            else:
                self.msg('p fold')

        #self.render()

    def act(self, to_call):
        to_call = int(to_call)
        if to_call == 0 and not self.game.flopped:
            self.msg('p check')

        #LOL AI
        if self.game.flopped and not self.folded:
            self.check_cards()

            if self.curr_rank > .7:
                self.msg('p raise ' + str(random.randint(50, 300)))

            elif self.curr_rank > .8:
                self.msg('p jam')

            elif self.curr_rank < .3:
                self.msg('p fold')
                self.folded = True;
                self.displaycards = []

            elif to_call == 0:
                if random.random() > .4:
                    self.msg('p raise' + str(random.randint(50, 300)))
                else:
                    self.msg('p check')

            else:
                if self.curr_rank > .5:
                    self.msg('p call')
                else:
                    self.msg('p fold')

        else:
            if random.random() > .8:
                self.msg('p fold')

            self.msg('p call')

        self.render()

    def render(self):
        self.surf.fill([255,255,255,0])
        if self.folded:
            return
        for i,d in enumerate(self.displaycards):
            d.draw(self.surf, (i*55,i*35))

        if self.id == 0:
            self.surf = pygame.transform.rotate(self.surf, -90)

        if self.id == 1:
            self.surf = pygame.transform.rotate(self.surf, -180)

        if self.id == 2:
            self.surf = pygame.transform.rotate(self.surf, 90)

    def draw(self, other):
        if self.busted:
            return

        white = [255,255,255,50]
        if self.game.player_to_act == self.name:
            white = [215,255,215,50]
        x, y = 0, 0
        if self.id == 0:
            x, y = (-1 *DisplayCard.h) + 150, (RESOLUTION_Y/2)-(DisplayCard.h/2)
            other.blit(self.surf, (x,y))
            x += 140
            y -= 86
            other.fill(white,(x, y, 120, 70))
            other.blit(self.nameplate, (x+10, y+10))
            other.blit(MISC_FONT.render(str(self.cash), True, [60,60,60]), (x+10, y+45))
            
        if self.id == 1:
            x, y = RESOLUTION_X/2 - (DisplayCard.w/2), (-1 * DisplayCard.h) + 100
            other.blit(self.surf, (x, y))
            y += 188
            x -= 130
            other.fill(white,(x, y, 120, 70))
            other.blit(self.nameplate, (x+10, y+10))
            other.blit(MISC_FONT.render(str(self.cash), True, [60,60,60]), (x+10, y+45))

        if self.id == 2:
            x, y = RESOLUTION_X-150, RESOLUTION_Y/2-(DisplayCard.h/2)
            other.blit(self.surf, (x, y))
            y -= 76
            x += 15
            other.fill(white,(x, y, 120, 70))
            other.blit(self.nameplate, (x+10, y+10))
            other.blit(MISC_FONT.render(str(self.cash), True, [60,60,60]), (x+10, y+45))

        

    def check_cards(self):
        b = [Card.new(x) for x in self.game.board]
        h = [Card.new(x) for x in self.hand]
        #print  b,h
        if len(b) < 3:
            return
        self.curr_rank = (7642 - self.ranker.evaluate(b,h)) / 7642.0
        ##print  self.name, "checked cards: ", self.curr_rank


class Player(Bot):
    def __init__(self, game):
        self.name = "player"
        Bot.__init__(self, game, 10, name=self.name)
        self.nameplate = MISC_FONT.render('human', True, [60,60,60])

        self.surf = Surface((DisplayCard.w*2 + 10, DisplayCard.h)).convert_alpha()

    def start(self):
        self.msg('p bankroll '+ str(START_CASH))
        self.msg('p start')

    def join(self):
        self.game.p.fake_players['player'] = self
        self.msg('p join')

    def render(self):
        self.surf.fill([255,255,255,0])
        if self.folded:
            return
        for i,d in enumerate(self.displaycards):
            d.facedown = False
            d.render() 
            d.draw(self.surf, (i*(DisplayCard.w + 10), 0))

    def draw(self, other):
        white = [255,255,255,50]
        if self.game.player_to_act == 'player':
            white = [215,255,215,50]
        x = (RESOLUTION_X/2-DisplayCard.w +15)
        y = RESOLUTION_Y-(DisplayCard.h-40)
        other.blit(self.surf, (x,y-3))
        y -= 86
        other.fill(white,(x, y, 120, 70))
        other.blit(self.nameplate, (x+10, y+10))
        other.blit(MISC_FONT.render(str(self.cash), True, [60,60,60]), (x+10, y+45))

    def act(self, to_call):
        self.render()
        pass

    def handle(self, msg):
        Bot.handle(self, msg)


   # def render(self):
   #     


if __name__ == '__main__':
    m = Main(SCREEN)
    m.run()